import datetime
import hashlib
import hmac
import urllib.parse

from lowhaio import (
    EmptyAsyncIterator,
)


def signed(request, credentials, service, region):

    async def _signed(method, url, params=(), headers=(),
                      body=EmptyAsyncIterator, body_args=(), body_kwargs=()):

        body_hash = hashlib.sha256()
        chunks = []
        length = 0

        # The body must be buffered to get a hash before the request is
        # initiated, but the chunks don't need to be concatanated together
        async for chunk in body(*body_args, **dict(body_kwargs)):
            body_hash.update(chunk)
            chunks.append(chunk)
            length += len(chunk)

        async def hashed_body():
            for chunk in chunks:
                yield chunk

        access_key_id, secret_access_key, auth_headers = await credentials()

        parsed_url = urllib.parse.urlsplit(url)
        all_headers = aws_sigv4_headers(
            access_key_id, secret_access_key,
            headers + auth_headers + ((b'content-length', str(length).encode()),), service, region,
            parsed_url.hostname, method.decode(), parsed_url.path, params, body_hash.hexdigest(),
        )

        return await request(method, url, params=params, headers=all_headers, body=hashed_body)

    return _signed


def aws_sigv4_headers(access_key_id, secret_access_key, pre_auth_headers,
                      service, region, host, method, path, params, body_hash):
    algorithm = 'AWS4-HMAC-SHA256'

    now = datetime.datetime.utcnow()
    amzdate = now.strftime('%Y%m%dT%H%M%SZ')
    datestamp = now.strftime('%Y%m%d')
    credential_scope = f'{datestamp}/{region}/{service}/aws4_request'

    pre_auth_headers_lower = tuple(
        (header_key.decode().lower(), ' '.join(header_value.decode().split()))
        for header_key, header_value in pre_auth_headers
    )
    required_headers = (
        ('host', host),
        ('x-amz-content-sha256', body_hash),
        ('x-amz-date', amzdate),
    )
    headers = sorted(pre_auth_headers_lower + required_headers)
    signed_headers = ';'.join(key for key, _ in headers)

    def signature():
        def canonical_request():
            canonical_uri = urllib.parse.quote(path, safe='/~')
            quoted_params = sorted(
                (urllib.parse.quote(key, safe='~'), urllib.parse.quote(value, safe='~'))
                for key, value in params
            )
            canonical_querystring = '&'.join(f'{key}={value}' for key, value in quoted_params)
            canonical_headers = ''.join(f'{key}:{value}\n' for key, value in headers)

            return f'{method}\n{canonical_uri}\n{canonical_querystring}\n' + \
                   f'{canonical_headers}\n{signed_headers}\n{body_hash}'

        def sign(key, msg):
            return hmac.new(key, msg.encode('ascii'), hashlib.sha256).digest()

        string_to_sign = f'{algorithm}\n{amzdate}\n{credential_scope}\n' + \
                         hashlib.sha256(canonical_request().encode('ascii')).hexdigest()

        date_key = sign(('AWS4' + secret_access_key).encode('ascii'), datestamp)
        region_key = sign(date_key, region)
        service_key = sign(region_key, service)
        request_key = sign(service_key, 'aws4_request')
        return sign(request_key, string_to_sign).hex()

    return (
        (b'authorization', (
            f'{algorithm} Credential={access_key_id}/{credential_scope}, '
            f'SignedHeaders={signed_headers}, Signature=' + signature()).encode('ascii')
         ),
        (b'x-amz-date', amzdate.encode('ascii')),
        (b'x-amz-content-sha256', body_hash.encode('ascii')),
    ) + pre_auth_headers
